#include <avr/io.h>
#define F_CPU	16000000UL
#include <util/delay.h>
#include <stdio.h>
#include <stdlib.h>
#include "lcd.h"

#define BUTTON_LEFT		PA0
#define BUTTON_OK		PA1
#define BUTTON_RIGHT	PA2
#define BUTTON_RST		PA3

#define BUTTON_DDR		DDRA
#define BUTTON_PIN		PINA
#define BUTTON_PORT		PORTA

#define LEFT 0
#define OK 1
#define RIGHT 2
#define RESET 3

#define PIXEL_OFF			0
#define PIXEL_ON			1
#define NOT_DEFINED			2

#define HORIZONTAL 1
#define VERTICAL 0

#define LCDCOLMAX	84
#define LCDROWMAX	6

#define WIDTH 9
#define HEIGHT 5
#define NR_COL 7
#define NR_ROW 6

#define FIRST_ROW -1
#define BLACK 0
#define WHITE 1
#define EMPTY 2

static unsigned char lcd_buffer[LCDROWMAX][LCDCOLMAX];

static unsigned char x_ship = 40; 

static unsigned char x_proj, y_proj = 41; 

static unsigned char x_enemy = 41, y_enemy = 5; 

static char x_e_step;
static unsigned char x_e_counter, y_e_counter;

static unsigned char enemymap[3][9]; 


static unsigned char enemies_alive;

static unsigned char lives = 51;


unsigned char battleship[5][5] = {{0,0,1,0,0},
						  {0,1,1,1,0},
						  {0,1,0,1,0},
						  {1,1,1,1,1},
						  {1,1,1,1,1}};


unsigned char projectile[5][5] = {{0,0,0,0,0},
						  {0,0,1,0,0},
						  {0,1,1,1,0},
						  {0,0,1,0,0},
						  {0,0,0,0,0}};
  
unsigned char enemy[5][5] = 	 {{0,0,1,0,0},
						  {0,0,1,0,0},
						  {0,1,1,1,0},
						  {0,0,1,0,0},
						  {0,1,0,1,0}};
						  

char eggs[9][2];

char potentialEggs[9];

/* displays a start screen. waits until the user presses a key */
void start_screen(void)
{
	lcd_clear();
	lcd_goto_xy_exact(0,0);
	lcd_goto_xy(3,2);
	lcd_chr(ARROW_LEFT);
	lcd_str("Chicken");
	
	lcd_goto_xy(4,3);
	lcd_str("Invaders");
	lcd_chr(ARROW_RIGHT);
	
	_delay_ms(100);

	while((PINA & (1 << PA0)) && (PINA & (1 << PA1)) && (PINA & (1 << PA2)));
}

void initEnemies(void)
{
	int i, j;
	
	x_enemy = 41;
	y_enemy = 5;
	x_e_counter = 0;
	x_e_step = 1;
	y_e_counter = 0;
	
	for(i = 0; i < 3; i++)
		for(j = 0; j < 9; j++)
			enemymap[i][j] = 1;
	
	enemies_alive = 27;	
}

/* Move text position cursor to specified position -> lcd library modified */
void gotoXY(unsigned char x, unsigned char y) {
    if (x > LCDCOLMAX - 1) 
		x = LCDCOLMAX - 1 ; // ensure within limits
    if (y > LCDROWMAX - 1) 
		y = LCDROWMAX - 1 ; // ensure within limits

	lcd_send(0x80 | x, LCD_CMD);   //column
	lcd_send(0x40 | y, LCD_CMD);   //row
}

/* Set a single pixel either on or off, update display buffer. -> lcd library modified */
void setpixel( unsigned char x, unsigned char y, unsigned char c ) {
	unsigned char value;
	unsigned char row;
	
	row = y / 8;

	value = lcd_buffer[row][x];
	
	if (c != PIXEL_OFF)
		value &= ~(1 << (y % 8));
	else
		value |= (1 << (y % 8));

	lcd_buffer[row][x] = value;
	
	gotoXY(x,row);
	lcd_send(value, LCD_DATA);
}

/* draws the enemies */
void drawEnemies(void)
{
	int i, j, x, y, k, l;
	
	for(i = 0; i < 3; i++)
		for(j = 0; j < 9; j++)
		{
			x = x_enemy - (4 - j) * 8;
			y = y_enemy + 8 * i;
			
			lcd_goto_xy_exact(0,0);
			
			if(enemymap[i][j] == 1)
				for(k = 0; k < 5; k++)
					for(l = 0; l < 5; l++)
						if(enemy[k][l] == 1)
							setpixel((int)x + l, (int)y + k, PIXEL_OFF);
		}
}

void drawBattleship(unsigned char x, unsigned char y)
{
	char i, j;
	
	lcd_goto_xy_exact(0,0);
	for(i = 0; i < 5; i++)
		for(j = 0; j < 5; j++)
			if(battleship[i][j] == 1)
				setpixel((int)x + j, (int)y + i, PIXEL_OFF);
	
}

void drawProjectile(unsigned char x, unsigned char y)
{
	char i, j;
	
	if (y_proj <= 40)
	{
		lcd_goto_xy_exact(0,0);
		for(i = 0; i < 5; i++)
			for(j = 0; j < 5; j++)
				if(projectile[i][j] == 1)
					setpixel((int)x + j, (int)y + i, PIXEL_OFF);
	}
}


/* reset lcd and update buf data */
void reset_screen(void)
{
	int i, j;
	lcdCacheIdx = 0;
	
	lcd_base_addr(lcdCacheIdx);
	
    // Set the entire cache to zero and write 0s to lcd
	for( i=0;i < LCD_CACHE_SIZE * 8;i++) {
		lcd_send(0, LCD_DATA);
    }
	
	for(i = 0; i < LCDROWMAX; i++)
		for(j = 0; j < LCDCOLMAX; j++)
		{
			lcd_buffer[i][j] = 0x00;
		}
}

void init_buttons(void)
{
	/* setare directie pini de iesire */
	BUTTON_DDR &= ~(1<<BUTTON_LEFT);
	BUTTON_DDR &= ~(1<<BUTTON_OK);
	BUTTON_DDR &= ~(1<<BUTTON_RIGHT);
	BUTTON_DDR &= ~(1<<BUTTON_RST);

	BUTTON_PORT |= (1<<BUTTON_LEFT); 
	BUTTON_PORT |= (1<<BUTTON_OK); 
	BUTTON_PORT |= (1<<BUTTON_RIGHT);
	BUTTON_PORT |= (1<<BUTTON_RST);
}

void move_ship(void)
{
	if (x_ship > 0)
		if (PINA & (1 << PA0))
			x_ship--;
			
	if (x_ship < 80)
		if (PINA & (1 << PA2))
			x_ship++;	
	
}

void launch_projectile(void)
{
	if (y_proj > 40)
		if (!(PINA & (1 << PA1)))
		{
			x_proj = x_ship;
			y_proj = 40;
		}
}

void move_projectile(void)
{
	if (y_proj <= 40)
		y_proj--;
}

void move_enemies(void)
{
	if(x_e_step > 0 && x_e_counter == 20)
		x_enemy += 2 * x_e_step;
	if(x_e_step > 0 && x_e_counter > 40)
	{
		x_enemy += 2 * x_e_step;
		x_e_step *= -1;
	}
	
	if(x_e_step < 0 && x_e_counter == 20)
		x_enemy += 2 * x_e_step;
	if(x_e_step < 0 && x_e_counter < 0)
	{
		x_enemy += 2 * x_e_step;
		x_e_step *= -1;
	}
	
	x_e_counter += x_e_step;
}

void collision(void)
{
	int i, j;
	
	if(y_proj <= y_enemy + 2 * 8)
	{
			for(j = 0; j < 3; j++)
			{
				for(i = 0; i < 9; i++)
					if(y_proj < y_enemy + 8 * j && y_proj > y_enemy + 8 * (j - 1))
						if(enemymap[j][i] != 0)
							if(x_proj > x_enemy - (4 - i) * 8 - 2 && x_proj < x_enemy - (4 - i) * 8 + 4)
							{
								enemymap[j][i] = 0;
								y_proj = 41;
								enemies_alive--;
								y_proj = 100;
							}
			}
	}
}

void livesScore(void) {
	lcd_goto_xy_exact(0,0);
	lcd_goto_xy(1,1);
	lcd_chr('L');
	lcd_goto_xy(1,2);
	lcd_chr('i');
	lcd_goto_xy(1,3);
	lcd_chr('v');
	lcd_goto_xy(1,4);
	lcd_chr('e');
	lcd_goto_xy(1,5);
	lcd_chr('s');
	lcd_goto_xy(1,6);
	lcd_chr(lives);
}

int moveEggs(int id) {
	
	if (eggs[id][0] == 0)
		eggs[id][0] = x_enemy - (4 - id) * 8;
	if (eggs[id][1] == 0)
		eggs[id][1] = y_enemy + 8;

	if (eggs[id][1] < 46) {
		setpixel(eggs[id][0], eggs[id][1]++, PIXEL_OFF);
		return 1;
	}
	else {
		eggs[id][0] = 0;
		eggs[id][1] = 0;
		return 0;
	}
}

void eggsCollision(void) {
	int i;
	
	for (i = 0; i < 9; i++)
		if (eggs[i][0] >= x_ship && eggs[i][0] - 5 <= x_ship
				&& eggs[i][1] > 40) {
			lives--;
			potentialEggs[i] = 0;
			eggs[i][0] = 0;
			eggs[i][1] = 0;
			livesScore();
			_delay_ms(1000);
		}
}

void gameOver(void) {
	lcd_goto_xy_exact(0,0);
	lcd_goto_xy(4,3);
	lcd_str("GAME OVER");
}

void finish(void) {
	lcd_goto_xy_exact(0,0);
	lcd_goto_xy(2,3);
	lcd_str("MISSION");
	lcd_goto_xy(3,4);
	lcd_str("COMPLETE");
}

int main(void) 
{	
	init_buttons();
	
	int j;
	
	/* Setup LCD */
 	lcd_init();
	lcd_contrast(0x40);
	reset_screen();
	start_screen();
	initEnemies();
	reset_screen();
	drawEnemies();
	
	int i = 0;
	
	while(1) {
		collision();
		
		livesScore();
	
		move_ship();
		drawBattleship(x_ship,40);
		
		launch_projectile();
		move_projectile();
		drawProjectile(x_proj, y_proj);
		
		drawEnemies();
		
		_delay_ms(200);
		reset_screen();
		
		i++;
		if (i % 16 == 0) {
			while (enemymap[0][i % 9] == 0)
				i++;
			potentialEggs[i % 9] = 1;
		}
		
		for (j = 0; j < 9; j++) {
			if (potentialEggs[j]) {
				if (moveEggs(j) == 0)
					potentialEggs[j] = 0;
				eggsCollision();
			}
		}
		
		if (lives == '0') {
			reset_screen();
			gameOver();
			break;
		}
		else if (enemies_alive == 0) {
			finish();
			break;
		}
		
		move_enemies();
	}

}